/**
 * Copyright (c) 2009 Carnegie Mellon University.
 * All rights reserved.
 *
 * Permission to use this software and its documentation for any purpose is
 * hereby granted, provided that the above copyright notice appear and that
 * both that copyright notice and this permission notice appear in supporting
 * documentation, and that the name of CMU not be used in advertising or
 * publicity pertaining to distribution of the software without specific,
 * written prior permission.
 *
 * CMU DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
 * IMPLIED WSTRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL CMU BE
 * LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, RISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * GCC is free software; you can redistribute it and/or modify it under|
 * the terms of the GNU General Public License as published by the Free|
 * Software Foundation; either version 3, or (at your option) any later|
 * version.|
 * |
 * GCC is distributed in the hope that it will be useful, but WITHOUT ANY|
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or|
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License|
 * for more details.|
 * |
 * You should have received a copy of the GNU General Public License|
 * along with GCC; see the file COPYING3.  If not see|
 * <http://www.gnu.org/licenses/>.
 */

#include "air-internal.h"
#include "air-errorcodes.h"

#define BUILD_CALL(insn) \
  sprintf (callbuf, "%s%s", flag_check_overflow_nop ? "nop" : \
                    sign == RTX_UNSIGNED ? "call\tconstraint_handler_wrapper_u" : \
                    "call\tconstraint_handler_wrapper_s", \
                    flag_check_overflow_nop ? "" : (insn))

#define OUTPUT_ASM_INSN(insn) \
  sprintf (insnbuf, insn, suffix); \
  output_asm_insn (insnbuf, operands)

static void
output_decrement (unsigned int sign, rtx *operands, const char *suffix)
{
  char insnbuf[100];
  char callbuf[100];
  BUILD_CALL ("dec");
  if (sign == RTX_SIGNED) {
    OUTPUT_ASM_INSN ("dec{%s}\t%%0");
    output_asm_insn ("jno\t.LANALYZE%=", NULL);
    output_asm_insn (callbuf, NULL);
    output_asm_insn (".LANALYZE%=", NULL);
  }
  else if (sign == RTX_UNSIGNED) {
    OUTPUT_ASM_INSN ("test{%s}\t%%0, %%0");
    output_asm_insn (insnbuf, operands);
    output_asm_insn ("jnz .LANALYZE%=", NULL);
    output_asm_insn (callbuf, NULL);
    output_asm_insn (".LANALYZE%=:", NULL);
    OUTPUT_ASM_INSN ("dec{%s}\t%%0");
  }
  else {
    OUTPUT_ASM_INSN ("dec{%s}\t%%0");
  }
}

static void
output_shift (rtx *operands, unsigned int sign)
{
  char callbuf[100];
  BUILD_CALL ("sal");
  if (CONST_INT_P(operands[2])) {
    int mask;
    char testbuf[100];
    unsigned int shift = INTVAL(operands[2]);
    if (shift == 0) return;
    mask = ((int)0x80000000) >> (shift - 1);
    sprintf (testbuf, "test{l}\t{$0x%x, %%0|%%0, $0x%x}", mask, mask);
    output_asm_insn (testbuf, operands);
    output_asm_insn ("jz\t.LANALYZE%=", NULL);
    output_asm_insn (callbuf, NULL);
    output_asm_insn (".LANALYZE%=:", NULL);     
    output_asm_insn ("sal{l}\t{%b2, %0|%0, %b2}", operands);
    return;
  }
  output_asm_insn ("test{b}\t%b2, %b2", operands);
  output_asm_insn ("jz\t.LANALYZE%=", NULL);
  output_asm_insn ("cmp{l}\t{$31, %k2|%k2, $31}", operands);
  output_asm_insn ("ja\t.LANALYZE_OVERFLOW%=", NULL);
  output_asm_insn ("dec\t%b2", operands);
  output_asm_insn ("or{l}\t{$0x80000000, %k2|%k2, $0x80000000}", operands);
  output_asm_insn ("sar{l}\t{%b2, %k2|%k2, %b2}", operands);
  output_asm_insn ("test{l}\t{%k2, %0|%0, %k2}", operands);
  output_asm_insn ("jz\t.LANALYZE_INNER%=", NULL);
  output_asm_insn (callbuf, NULL);
  output_asm_insn (".LANALYZE_INNER%=:", NULL);
  output_asm_insn ("bsf\t%k2, %k2", operands);
  output_asm_insn ("neg{b}\t%b2", operands);
  output_asm_insn ("add{b}\t{$0x20, %b2|%b2, $0x20}", operands);
  output_asm_insn ("jmp\t.LANALYZE_NOOVERFLOW%=", NULL);     
  output_asm_insn (".LANALYZE_OVERFLOW%=:", NULL);     
  output_asm_insn (callbuf, NULL);
  output_asm_insn (".LANALYZE_NOOVERFLOW%=:", NULL);     
  output_asm_insn ("sal{l}\t{%b2, %0|%0, %b2}", operands);
  output_asm_insn (".LANALYZE%=:", NULL);     
}

/* Output the jump/call/label assembly code necessary to check for
   overflow given that the previous operation used operations with
   the given sign.  */

static void
output_overflow_op (unsigned int sign, opcode_type_t opcode)
{
  if (sign == RTX_WHITELIST) {
    output_asm_insn ("# Whitelisted operation", NULL);
    return;
  }
  if (sign == RTX_UNSIGNED) {
    /* The carry flag is not changed by increment instructions.
    ** The zero flag must be checked to determine whether wrapping
    ** occurred.
    */
    if (opcode == OPCODE_TYPE_INC)
      output_asm_insn ("jnz\t.LANALYZE%=", NULL);
    else
      output_asm_insn ("jnc\t.LANALYZE%=", NULL);
  }
  else if (sign == RTX_SIGNED || opcode == OPCODE_TYPE_MUL)
    output_asm_insn ("jno\t.LANALYZE%=", NULL);
  else {
    output_asm_insn ("# No sign information attached", NULL);
    return;
  }

  if (flag_check_overflow_nop) {
    output_asm_insn ("nop", NULL);
  }
  else {
    switch (opcode) {
      case OPCODE_TYPE_ADD:
        output_asm_insn (sign == RTX_UNSIGNED ?
            "call\tconstraint_handler_wrapper_uadd" :
            "call\tconstraint_handler_wrapper_sadd",
            NULL);
        break;
      case OPCODE_TYPE_SUB:
        output_asm_insn (sign == RTX_UNSIGNED ?
            "call\tconstraint_handler_wrapper_usub" :
            "call\tconstraint_handler_wrapper_ssub",
            NULL);
        break;
      case OPCODE_TYPE_MUL:
        output_asm_insn (sign == RTX_UNSIGNED ?
            "call\tconstraint_handler_wrapper_umul" :
            "call\tconstraint_handler_wrapper_smul",
            NULL);
        break;
      case OPCODE_TYPE_INC:
        output_asm_insn (sign == RTX_UNSIGNED ?
            "call\tconstraint_handler_wrapper_uinc" :
            "call\tconstraint_handler_wrapper_sinc",
            NULL);
        break;
      case OPCODE_TYPE_NEG:
        output_asm_insn (sign == RTX_UNSIGNED ?
            "call\tconstraint_handler_wrapper_uneg" :
            "call\tconstraint_handler_wrapper_sneg",
            NULL);
        break;
      default:
        gcc_unreachable ();
        break;
    }
  }
  output_asm_insn (".LANALYZE%=:", NULL);
}

/* Extract the sign of the operation based on the sign of its
   operands. Revert to RTX_NO_SIGN if the operation is stack
   arithmetic or if the sign is unsigned, but no unsigned
   overflow ias being checked for.  */

static unsigned int
get_sign (rtx op0, rtx op1)
{
  char buf[100];
#ifdef AIR_DEBUG
  fprintf (stderr, "---------- In get_sign ----------\n");
  debug_rtx (op0);
  if (op1)
    debug_rtx (op1);
#endif

  if (!op1) {
    unsigned int sign = op0->whitelisted ? RTX_WHITELIST : RTX_SIGN (op0);
    sprintf (buf, "# (%d %s)", sign, GET_RTX_NAME (GET_CODE (op0)));
    output_asm_insn (buf, NULL);

    if (RTX_FLAG (op0, frame_related) ||
        (flag_overflow_level == 1 && sign == RTX_UNSIGNED))
      return RTX_NO_SIGN;
    return sign;
  }
  else {
    unsigned int sign1 = op0->whitelisted ? RTX_WHITELIST : RTX_SIGN (op0);
    unsigned int sign2 = CONST_INT_P (op1) ? RTX_NO_SIGN :
                         op1->whitelisted ? RTX_WHITELIST : RTX_SIGN (op1);
    sprintf (buf, "# (%d %s) (%d %s)", sign1, GET_RTX_NAME (GET_CODE (op0)),
        sign2, GET_RTX_NAME (GET_CODE (op1)));
    output_asm_insn (buf, NULL);
    if (RTX_FLAG (op0, frame_related) ||
        RTX_FLAG (op1, frame_related))
      return RTX_NO_SIGN;
    if (sign1 == RTX_WHITELIST ||
        sign2 == RTX_WHITELIST)
      return RTX_WHITELIST;
    else if (sign1 == RTX_UNSIGNED ||
        sign2 == RTX_UNSIGNED)
    {
      if (flag_overflow_level == 1)
        return RTX_NO_SIGN;
      return RTX_UNSIGNED;
    }
    else if (sign1 == RTX_SIGNED ||
        sign2 == RTX_SIGNED)
      return RTX_SIGNED;
    return RTX_NO_SIGN;
  }
}

#define COPY_SIGN(x) \
  if (RTX_SIGN (operands[0]) == RTX_NO_SIGN) \
    RTX_SIGN (operands[1]) = RTX_SIGN (operands[x])

/* Generate and output the assembly code for the addsi_1 insn
   with overflow checks afterward.  */

void
air_add_1 (rtx insn, rtx *operands, const char *suffix)
{
  opcode_type_t type = OPCODE_TYPE_UNKNOWN;
  unsigned int sign;
  char insnbuf[100];
  switch (get_attr_type (insn))
  {
    case TYPE_LEA:
      if (CONST_INT_P (operands[2])
          && (INTVAL (operands[2]) < 0)) {
        operands[2] = GEN_INT (-INTVAL (operands[2]));
        if (rtx_equal_p (operands[0], operands[1])) {
          COPY_SIGN (1);
          sign = get_sign (operands[0], operands[2]);
          OUTPUT_ASM_INSN ("sub{%s}\t{%%2, %%0|%%0, %%2}");
        }
        else if (rtx_equal_p (operands[0], operands[2])) {
          COPY_SIGN (2);
          sign = get_sign (operands[0], operands[1]);
          OUTPUT_ASM_INSN ("sub{%s}\t{%%1, %%0|%%0, %%1}");
        }
        else {
          COPY_SIGN (1);
          OUTPUT_ASM_INSN ("mov{%s}\t{%%1, %%0|%%0, %%1}");
          sign = get_sign (operands[0], operands[2]);
          OUTPUT_ASM_INSN ("sub{%s}\t{%%2, %%0|%%0, %%2}");
        }
        type = OPCODE_TYPE_SUB;
      }
      else {
        if (rtx_equal_p (operands[0], operands[1])) {
          COPY_SIGN (1);
          sign = get_sign (operands[0], operands[2]);
          OUTPUT_ASM_INSN ("add{%s}\t{%%2, %%0|%%0, %%2}");
        }
        else if (rtx_equal_p (operands[0], operands[2])) {
          COPY_SIGN (2);
          sign = get_sign (operands[0], operands[1]);
          OUTPUT_ASM_INSN ("add{%s}\t{%%1, %%0|%%0, %%1}");
        }
        else {
          COPY_SIGN (1);
          OUTPUT_ASM_INSN ("mov{%s}\t{%%1, %%0|%%0, %%1}");
          sign = get_sign (operands[0], operands[2]);
          OUTPUT_ASM_INSN ("add{%s}\t{%%2, %%0|%%0, %%2}");
        }
        type = OPCODE_TYPE_ADD;
      }
      break;
    case TYPE_INCDEC:
      gcc_assert (rtx_equal_p (operands[0], operands[1]));
      if (operands[2] == const1_rtx) {
        COPY_SIGN (1);
        sign = get_sign (operands[0], NULL);
        OUTPUT_ASM_INSN ("inc{%s}\t%%0");
        type = OPCODE_TYPE_INC;
      }
      else {
        gcc_assert (operands[2] == constm1_rtx);
        COPY_SIGN (1);
        sign = get_sign (operands[0], NULL);
        output_decrement (sign, operands, suffix);
        return;
      }
      break;
    default:
      /* Use add as much as possible to replace lea for AGU optimization. */
#ifdef TAARGET_OPT_AGU
      if (which_alternative == 2 && TARGET_OPT_AGU) {
        sign = get_sign (operands[0], operands[1]);
        OUTPUT_ASM_INSN ("add{%s}\t{%%1, %%0|%%0, %%1}");
        type = OPCODE_TYPE_ADD;
        break;
      }
#endif

      gcc_assert (rtx_equal_p (operands[0], operands[1]));
      /* Make things pretty and `subl $4,%eax' rather than `addl $-4, %eax'. */
      if (CONST_INT_P (operands[2])
          /* Avoid overflows.  */
          && ((INTVAL (operands[2]) & ((((unsigned int) 1) << 31) - 1)))
          && (INTVAL (operands[2]) == 128
              || (INTVAL (operands[2]) < 0
                  && INTVAL (operands[2]) != -128))) {
        operands[2] = GEN_INT (-INTVAL (operands[2]));
        COPY_SIGN (1);
        sign = get_sign (operands[0], operands[2]);
        OUTPUT_ASM_INSN ("sub{%s}\t{%%2, %%0|%%0, %%2}");
        type = OPCODE_TYPE_SUB;
      }
      else {
        COPY_SIGN (1);
        sign = get_sign (operands[0], operands[2]);
        OUTPUT_ASM_INSN ("add{%s}\t{%%2, %%0|%%0, %%2}");
        type = OPCODE_TYPE_ADD;
      }
      break;
  }
  output_overflow_op (sign, type);
}

/* Generate and output the assembly code for the addsi_2 insn
   with overflow checks afterward.  */

void
air_add_2(rtx insn, rtx *operands, const char *suffix)
{
  opcode_type_t type = OPCODE_TYPE_UNKNOWN;
  unsigned int sign;
  char insnbuf[100];
  switch (get_attr_type (insn))
  {
    case TYPE_INCDEC:
      gcc_assert (rtx_equal_p (operands[0], operands[1]));
      COPY_SIGN (1);
      if (operands[2] == const1_rtx) {
        sign = get_sign (operands[0], NULL);
        OUTPUT_ASM_INSN ("inc{%s}\t%%0");
        type = OPCODE_TYPE_INC;
      }
      else {
        gcc_assert (operands[2] == constm1_rtx);
        sign = get_sign (operands[0], NULL);
        output_decrement (sign, operands, suffix);
        return;
      }
      break;
    default:
      gcc_assert (rtx_equal_p (operands[0], operands[1]));
      COPY_SIGN (1);
      /* Make things pretty and `subl $4,%eax' rather than `addl $-4, %eax'.
         Exceptions: -128 encodes smaller than 128, so swap sign and op.  */
      if (CONST_INT_P (operands[2])
          /* Avoid overflows.  */
          && ((INTVAL (operands[2]) & ((((unsigned int) 1) << 31) - 1)))
          && (INTVAL (operands[2]) == 128
              || (INTVAL (operands[2]) < 0
                  && INTVAL (operands[2]) != -128))) {
        operands[2] = GEN_INT (-INTVAL (operands[2]));
        sign = get_sign (operands[0], operands[2]);
        OUTPUT_ASM_INSN ("sub{%s}\t{%%2, %%0|%%0, %%2}");
        type = OPCODE_TYPE_SUB;
      }
      else {
        sign = get_sign (operands[0], operands[2]);
        OUTPUT_ASM_INSN ("add{%s}\t{%%2, %%0|%%0, %%2}");
        type = OPCODE_TYPE_ADD;
      }
      break;
  }
  output_overflow_op (sign, type);
}

/* Generate and output the assembly code for the mulsi3_1 insn
   with overflow checks afterward.  */

void
air_mul3_1 (rtx *operands, const char *suffix)
{
  unsigned int sign;
  char insnbuf[100];
  if (which_alternative == 2) {
    sign = get_sign (operands[0], operands[2]);
    OUTPUT_ASM_INSN ("imul{%s}\t{%%2, %%0|%%0, %%2}");
  }
  else {
    sign = get_sign (operands[0], operands[1]);
    OUTPUT_ASM_INSN ("imul{%s}\t{%%2, %%1, %%0|%%0, %%1, %%2}");
  }
  output_overflow_op (sign, OPCODE_TYPE_MUL);
}

/* Generate and output the assembly code for the negsi2_1 insn
   with overflow checks afterward.  */

void
air_negsi2_1 (rtx *operands)
{
  output_asm_insn ("neg{l}\t%0", operands);
  output_overflow_op (get_sign (operands[0], NULL), OPCODE_TYPE_NEG);
}

/* Generate and output the assembly code for the ashlsi3_1 insn
   with overflow checks afterward.  */

void
air_ashlsi3_1 (rtx insn, rtx *operands)
{
  unsigned int sign = get_sign (operands[0], NULL);
  switch (get_attr_type (insn))
  {
    case TYPE_ALU:
      gcc_assert (operands[2] == const1_rtx);
      gcc_assert (rtx_equal_p (operands[0], operands[1]));
      output_asm_insn ("add{l}\t%0, %0", operands);
      output_overflow_op (sign, OPCODE_TYPE_ADD);
      break;
    case TYPE_LEA:
      /* The cost of a lea instruction should be set high enough
         to make this case impossible.  */
      gcc_unreachable();
      break;
    default:
      if (sign != RTX_SIGNED && sign != RTX_UNSIGNED) {
        if (REG_P (operands[2]))
          output_asm_insn ("sal{l}\t{%b2, %0|%0, %b2}", operands);
        else if (operands[2] == const1_rtx
            && (TARGET_SHIFT1 || optimize_function_for_size_p (cfun)))
          output_asm_insn ("sal{l}\t%0", operands);
        else
          output_asm_insn ("sal{l}\t{%2, %0|%0, %2}", operands);
      }
      else
        output_shift (operands, sign);
  }
}

/* Generate and output the assembly code for the subsi_1 insn
   with overflow checks afterward.  */

void
air_sub_1 (rtx *operands, const char *suffix)
{
  char insnbuf[100];
  unsigned int sign = get_sign (operands[0], operands[2]);
  OUTPUT_ASM_INSN ("sub{%s}\t{%%2, %%0|%%0, %%2}");
  output_overflow_op (sign, OPCODE_TYPE_SUB);
}

/* Generate and output the assembly code for the subsi_1 insn
   with overflow checks afterward.  */

void
air_add3_carry (rtx *operands, const char *suffix)
{
  char insnbuf[100];
  unsigned int sign = get_sign (operands[0], operands[2]);
  OUTPUT_ASM_INSN ("adc{%s}\t{%%2, %%0|%%0, %%2}");
  output_overflow_op (sign, OPCODE_TYPE_ADD);
}

