#!/usr/bin/perl
# 
# Copyright (c) 2009 Carnegie Mellon University.
# All rights reserved.
#
# Permission to use this software and its documentation for any purpose is
# hereby granted, provided that the above copyright notice appear and that
# both that copyright notice and this permission notice appear in supporting
# documentation, and that the name of CMU not be used in advertising or
# publicity pertaining to distribution of the software without specific,
# written prior permission.
#
# CMU DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
# IMPLIED WSTRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL CMU BE
# LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, RISING OUT OF OR
# IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#

# Intercept a air-instrumented program's output and accumulate the constraint
# violations encountered in a file. Running the program repeatedly with
# different arguments and using the same error file allows a large data set of
# errors to be collected incrementally. Used in conjunction with print-vuls.pl
# to print the error file.
#
# Sample usage: PROG ARGS 2>&1| ./catch-vuls.pl ERROR_FILE


$file = shift;

while (<STDIN>) {
  $line = $_;
  if ($line =~ /\t(.*?) at address 0x([0-9a-f]{7,8})/i) {
    push @addr, "$2 $1\n";
  }
}
@new_addr = sort(@addr);

open (FILE, "<$file");
@old_addr = <FILE>;
close FILE;

$new_cnt = 0;
$old_cnt = 0;

open (FILE, "+>$file");

while (($new_cnt < @new_addr) || ($old_cnt < @old_addr)) {
  if ($new_cnt == @new_addr) {
    $next = "$old_addr[$old_cnt++]";
  }
  elsif (($old_cnt == @old_addr) || ($new_addr[$new_cnt] le $old_addr[$old_cnt])) {
    $next = "$new_addr[$new_cnt++]";
  }
  else {
    $next = "$old_addr[$old_cnt++]";
  }

  if ($next ne $prev) {
    print FILE $next;
    $prev = $next;
  }
}

close FILE;

